#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihm.c,v $	GIPS $Revision: 2.2 $ (MIT/CSR) $Date: 1998/05/13 01:22:44 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		ihm.c		Creation Date:	01/31/84
**	Version:	$Revision: 2.2 $		Update:		$Date: 1998/05/13 01:22:44 $
**	Components:	main()		UNIX command
**			get_table()	save header details about file
**			get_poly()	save dscale=p coefficients
**			get_comp()	qsort() element comparison
**			get_logic()	decode -l option
**			get_map()	construct real <-> int map table
**			get_rasters()	construct x-y raster tables
**			get_rast()	construct a raster table
**			get_spool()	write input to /tmp
**			get_rsort()	sort input raster pointers
**			get_wrap()	fix up wrapped x-coordinates
**			get_debug()	write RASTER and MAP to 'stderr'
**
**			put_rast()	locate and process a raster
**			put_conv()	convert raster to float
**			put_poly()	convert dscale=p raster
**			put_pix()	select data from a raster
**			put_qpix()	quick select data from a raster
**			put_logic()	perform boolean pixel operation
**			cleanup()	cleanup routine
**
**	Function:	Merges or resamples one or more General Images
**
**	Method:		The code divides cleanly into two parts, and the
**			routines are named accordingly. In the "get" phase,
**	one or more General Image headers are read and certain variable values
**	saved in the chain of "table" structures. The output file is also
**	described in the same way (except that its header is written, not
**	read). Each axis (first x, then y) of each file is defined by a "map"
**	structure containing one index and one weight for each pixel boundary
**	on that axis (i.e. one more than the number of pixels). The "map"
**	elements are stored by increasing realworld coordinate value. The
**	weights are the widths of those pixels in realworld units. The crucial
**	step of the "get" phase occurs in get_rast() when the "map" tables of
**	input and output files are compared. This results in a "rast"
**	structure which determines the location within a given input file of
**	the pixels that may contribute to a given output pixel.
**
**	get_rast() can also determine whether an input image is to be read from
**	front-to-back, back-to-front, or essentially randomly. If the file
**	is not disk-resident and if it is not to be read sequentially,
**	get_spool() is called to copy it to a temporary disk, from which it
**	may be read randomly with lseek().
**
**	The "put" phase writes the output a raster at a time. The y-rast table
**	of each input file determines which input rasters contribute.  These
**	are read by put_rast(), converted to float by put_conv(), and their
**	contribution to the current output raster is computed in put_pix(),
**	using the appropriate x-rast table. put_rast() also buffers the input
**	rasters as far as it is able - not as raw input records but as "dwt"
**	structures output from put_pix(). This saves much time when ihm() is
**	resampling to reduce pixel resolution.
**
**	Structures:
**
**	MAP: Mapping between raster indices and pixels - one is allocated
**	by get_map() for each input and output image file. "mptr" are
**	indices to "physical" x- or y-coordinates (i.e. offsets to the
**	pixel fields in the files themselves). Initially, "mwt"
**	contains the realworld coordinate values corresponding to
**	"mptr". get_map() sorts the "map" elements in ascending "mwt"
**	value.
**
**	The tables are first used in "get_rast" to compare input and
**	output axis mappings to determine the relation between each
**	output pixel and contributing input pixels. Once "map" has
**	been used to construct a "rast" table, "mwt" become data
**	intervals (i.e. consecutive values are differenced) and are
**	used in the "put" phase as weights. Then the "mptr" indices
**	serve to locate input pixels. (Remember: there is no guarantee
**	that the input -> output mapping is 1 : 1).
**
**	RASTER: Table mapping between input and output axes -
**	there is one of these for each output pixel of each axis of
**	each input image. The entries map a given OUTPUT pixel to one
**	or more pixels of a particular input image. lowlimit and high-
**	limit are pointers to elements of the corresponding "map" and
**	"dwt" tables. lowweight and highweight are the weights for the
**	first and last pixels of the series, which will, in general,
**	only partly overlap the domain of the input pixel.
**	The remaining weights are taken from the "dwt" table.
**
**	DWT: general internal data & weight structure - the input rasters
**	read and saved by put_rast() are in the form of arrays of "dwt".
**	The output raster is kept in this form until converted to dtype.
**
**	TABLE: General Image descriptor table - one of these is generated
**	by get_table() to describe each input (and output) file. Although
**	not allocated first, "out->table" becomes the anchor for a chain
**	of "table" structures.
*/

#include	"ihm.h"
	/*
 	 *	variables that may be changed by input flags and arguments
 	 */
char	*ddir	= "/tmp";	/* spooling directory */
char	*logic	= NULL;		/* math logic stuff */
char	*ofile	= "-";		/* output filename */
flag	qflag;			/* true for "quick" merge */
flag	Dflag;			/* debug flag */
flag	qqflag = 1;		/* true for even quicker merge */

static	char	*hfile	= "-";	/* output header origin file */
static	char	*mfile	= "-";	/* secondary header origin file */
static	flag	Sflag;		/* generate memory statistics */
	/*
	 *	table used by gipsflags()
	 */
static	flagentry	flagtable[] = {
	"d",	"-> dir",	(flag *) &ddir,
	"h",	"-> file",	(flag *) &hfile,
	"l",	"-> logic",	(flag *) &logic,
	"m",	"-> file",	(flag *) &mfile,
	"o",	"-> file",	(flag *) &ofile,
	"q",	"1",		(flag *) &qflag,
	"D",	"1",		(flag *) &Dflag,
	"S",	"1",		(flag *) &Sflag,
	NULL,	"[file...]",	NULL
};

DWT	*accbuf;		/* buffer to accumulate output raster */
TABLE	*out;			/* anchor for i/o stream tables */
char	*inbuf;			/* common input buffer */
float	*convbuf;		/* common raster conversion buffer */
char	*outbuf;		/* output binary raster */
int	maxraster;		/* maximum input raster length */
int	maxconv;		/* maximum float raster length */

extern	TABLE	*get_table();	/* return pointer to new table */
extern	char	*index();	/* C library */
/*
 *	normal completion, or error, or interrupt signalled
 */
void
cleanup()
{
	if (out != NULL)
		while ((out = out->nexttab) != NULL)
			if (out->ftemp)
				(void)(unlink(out->fname));
			else if (out->noseek)
				while (getc(out->iop) != EOF)
					;
	exit(0);
}
main(argc,argv)

	int	argc;		/* argument count */
	char	**argv;		/* array of argument pointers */
{
	FILE	*fp;		/* merged header stream pointer */
	int	y;		/* y-raster index */
	register TABLE	*t;	/* pointer to input table */
	/*
	 *	save and validate callers arguments
	 */
	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	if (--argc == 0)
		argv[++argc] = "-";
	if (Sflag)
		telltale(0);
	/*
	 *	read the first input file header; allocate and fill the
	 *	first "TABLE" structure. This gets special treatment because
	 *	its header also serves as the default output header.
	 */
	t = get_table(*++argv);
	/*
	 *	"-h" option causes a fresh header file to be read. This may
	 *	be further augmented by the "-m" option.
	 */
	if (*hfile != '-')
		if ((fp = ihread(hfile,GIPSPROFILE)) == NULL)
			gipserr();
		else
			(void)(ihbcl(fp));
	if (*mfile != '-')
		if ((fp = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void)(ihbcl(fp));
	/*
	 *	add a history item to the incore header, and write it to 
	 *	the output file. get_table() with NULL argument denotes output
	 *	rather than input. A "TABLE" structure is allocated for the
	 *	output file, and is put before the "table" describing the
	 *	first input file. get_rasters() then compares the output 
	 *	header with that of the first input file and constructs "RASTER"
	 *	tables.
	 */
	putargs();
	get_rasters((out = get_table((char *)NULL))->nexttab = t);
	/*
	 *	get_table() and get_rasters() then process the remaining
	 *	input files, chaining their "table" blocks to the end of
	 *	the "out" list.
	 */
	while (--argc > 0)
		get_rasters(t = t->nexttab = get_table(*++argv));
	if (logic && *logic)
		error("too many -l fields: %s",logic);
	/*
	 *	up to four buffers are allocated:
	 *
	 *	inbuf   - into which the raw input rasters are read
	 *	convbuf - into which the float version of inbuf is written
	 *	accbuf  - in which weighted data and weights are accumulated
	 *		  for the current output raster.
	 *	outbuf	- the actual binary output raster
	 */
	if (! qflag)
		inbuf	= alloc(1+maxraster/sizeof(double),sizeof(double));
	accbuf = (DWT *) alloc(out->num[0],sizeof(struct dwt));
	convbuf	= (float *) alloc(maxconv,sizeof(float));
	outbuf	= alloc(out->num[0],out->dlen);
	/*
	 *	this is the "put" phase:
	 *
	 *	looping over output y-rasters, and by input file for each raster
	 */
	for (y = 0; y < out->num[1]; y++) {
		char	*c;		/* output pointer */

		clear((char *)accbuf,out->num[0] * sizeof(struct dwt));

		if (qqflag)
			clear(outbuf,out->num[0] * out->dlen);
		/*
		 *	the following loop over input files shows the way the
		 *	input rasters are selected and weighted using the "rast"
		 *	tables: tr points to the rast[1] element for the current
		 *	output y-value. The y-values of the input rasters that
		 *	contribute are given by elements tr->lowlimit through
		 *	tr->highlimit of the corresponding map[1] table.
		 */
		for (t = out; (t = t->nexttab) != NULL; ) {
			register RASTER * tr;
			register MAP 	* tm;

			tr = t->rast[1] + y;
			if ((tm = tr->lowlimit) == NULL)
				continue;
			if (t->spl[1] == -1) {
				while (++tm < tr->highlimit)
					;
				if (tm == tr->highlimit)
					put_rast(t,tm->mptr,tr->highweight);
				while (--tm > tr->lowlimit)
					put_rast(t,tm->mptr,tm->mwt);
				put_rast(t,tm->mptr,tr->lowweight);
			} else {
				put_rast(t,tm->mptr,tr->lowweight);
				while (++tm < tr->highlimit)
					put_rast(t,tm->mptr,tm->mwt);
				if (tm == tr->highlimit)
					put_rast(t,tm->mptr,tr->highweight);
			}
		}
		/*
		 *	the output pixels and their weights must now be
		 *	converted to the dtype of the output file. If the
		 *	weight is zero, the pixel is set to "dnull".
		 */
		t = out;
		if (qqflag)
			c = outbuf;
		else if (t->dtype == 'c')
			c = (char *)accbuf;
		else {
			register float	*buf	= convbuf;
			register	i;
			register DWT	*dw	= accbuf;

			for (i = t->num[0]; --i >= 0; dw++)
				if (dw->weight == 0.0)
					*buf++ = t->dnull;
				else {
					*buf++ = qflag ? dw->data :
						dw->data/dw->weight;
					if (t->dscale == 'n') {
						*--buf -= t->dzero;
						*buf++ *= t->dfact2;
					}
				}
			gconv(c = outbuf,(char *)convbuf,t->num[0],
				(int)ihdcvt(t->dtype),CVT_F);
		}
		if (fwrite(c,t->raster,1,t->iop) != 1)
			gipserr2(GIPSBADWRITE,ofile);
	}
	if (Sflag)
		telltale(1);
	(void)(cleanup());
}
