#ifndef lint
static char *sccsid = "%Z%$RCSfile: bidr.c,v $	GIPS $Revision: 1.14 $ (MIT/CSR) $Date: 2004/11/29 18:51:20 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1990 Massachusetts Institute of Technology
**
**	Module:		bidr.c		Creation Date:	07/22/90
**	Version:	$Revision: 1.14 $		Update:		$Date: 2004/11/29 18:51:20 $
**
**	Components:	gopen()		Access general image in cell format
**			gcell()		Return pointer to pixel
**			ginterp()	Return pointer to interpolation
**			gclose()	Close and release image
**			gwrite()	Mark image for update
**
**	Function:	Open and access general image in cell order
*/

#include	<stdio.h>
#include	<strings.h>
#define		GRID
#include	"gips.h"
#define		BIDR_MAGIC	0x7007

#define	ORBIT_IS_RIGHT_LOOKING(n)	 \
	(((n) >= 2219 && (n) <= 2464) || \
	 ((n) >= 2587 && (n) <= 2673) || \
	 ((n) >= 2682 && (n) <= 3667))

#define	uchar		unsigned char

	/*
	**	VAX binary conversions
	*/

#ifdef vax
#define	load_short(c)	(*(short *)(c))
#define	load_long(c)	(*(long *)(c))
#define	load_float(f,c)	(f)= *(float *)(c)
#else
#define	load_short(c)	((((uchar *)(c))[1]<<8)|((uchar *)(c))[0])
#define	load_long(c)	((((uchar *)(c))[3]<<24)|(((uchar *)(c))[2]<<16)|\
			(((uchar *)(c))[1]<<8)|((uchar *)(c))[0])
#define	load_float(f,c)	if (((uchar *)(c))[1]) *(unsigned long *)(&(f))= \
			(((((uchar *)(c))[1]-1)<<24)|(((uchar *)(c))[0]<<16)|\
				(((uchar *)(c))[3]<<8)|((uchar *)(c))[2]); \
			else (f)=0
#endif

	/*
	**	Image block descriptor
	*/

typedef	struct	bg	{
	struct	bg	*bg_next;	/* -> next image block */
	struct	bg	*bg_last;	/* -> last image block */
	int	bg_xorg;		/* x address to top-left pixel */
	int	bg_yorg;		/* y address to top-left pixel */
	int	bg_xnum;		/* number of columns */
	int	bg_ynum;		/* number of rows */
	float	bg_lat;			/* latitude of top-left pixel */
	float	bg_lon;			/* longitude of top-left pixel */
	off_t	bg_off;			/* offset in BIDR file */
} BG;

	/*
	**	Image file descriptor
	*/

typedef	struct	bi	{
	short	bi_magic;		/* BIDR_MAGIC */
	short	bi_flag;		/* see below */
#define		BI_EOF		01	/* EOF on BIDR */
#define		BI_DEBUG	02	/* Debug requested */
#define		BI_OFF4		04	/* p1,p2 offsets too large by 4 */
#define		BI_OFF1		010	/* p2 offset too large by 1 */
	char	*bi_fname;		/* BIDR file name */
	char	*bi_buf;		/* -> in-core image data block */
	int	bi_fid;			/* BIDR file descriptor */
	int	bi_xmax;		/* highest x address read */
	int	bi_xmin;		/* x addr of top left of 1st block */
	int	bi_ymax;		/* highest y address read */
	int	bi_ymin;		/* y addr of top left of 1st block */
	int	bi_size;		/* number of bytes in image block */
	int	bi_p1;			/* -> start of 1st raster in bi_buf */
	int	bi_p2;			/* -> end of 1st raster in bi_buf */
	BG	*bi_save;		/* -> BG of bi_buf */
	BG	*bi_first;		/* -> head of BG chain */
	BG	*bi_ptr;		/* -> current location in BG chain */
	BG	*bi_last;		/* -> tail of BG chain */
} BI;

	/*
	**	BIDR image header
	*/

static	struct	{
	char	h_sfdu[20];		/* NJPL label */
	uchar	h_ltype[2];		/* Secondary label type */
	uchar	h_llength[2];		/* Secondary label length */
	uchar	h_orbit[2];		/* Orbit number */
	uchar	h_dclass;		/* Data class */
	uchar	h_dlength;		/* Annotation label length */
	uchar	h_lines[2];		/* Number of image lines */
	uchar	h_pixels[2];		/* Number of bytes per line */
	uchar	h_olat[4];		/* Reference origin latitude */
	uchar	h_olon[4];		/* Reference origin longitude */
	uchar	h_rlat[4];		/* Reference latitude */
	uchar	h_rlon[4];		/* Reference longitude */
	uchar	h_ref_lines[4];		/* Reference offset lines */
	uchar	h_ref_pixels[4];	/* Reference offset samples */
	uchar	h_nburst[4];		/* Burst counter (F-BIDR only) */
	char	h_navid[32];		/* NAV unique ID (F-BIDR only) */
} hdr;

static	float	olon, olat;		/* origin latitude/longitude */
static	char	nulls[1024];		/* dnull bytes */
static	int	debug;			/* debug during open */
extern	int	errno;			/* system error code */
extern	int	_gcount;		/* raster valid byte count */
extern	char	*bidr_cell();		/* return pointer to raster */
static		read_bidr_index();

	/*
	**	Open the BIDR -- called from gopen()
	*/

gimage	*
bidr_open(gstr)
	char		*gstr;		/* geometry string */
{
	BI		*bi;		/* image structure pointer */
	char		buf[256];	/* scratch */
	double		pixsiz;		/* size of pixel in km */
	int		orbit;		/* image orbit number */
	int		i;		/* scratch */

	/*
	**	Get the BIDR file name, allocate a BI structure,
	**	open the BIDR
	*/

	if (sscanf(gstr, "type=b bidr=%s debug=%d", buf, &debug) != 2)
		return NULL;
	gipserrno = 0;
	bi = (BI *)alloc(sizeof(BI), 1);
	bi->bi_fname = strcpy(alloc(strlen(buf)+1, 1), buf);
	if ((bi->bi_fid = open(bi->bi_fname, 0)) == EOF) {
		(void)bidr_close((gimage *)bi);
		gipserrno = GIPSBADOPEN;
		return NULL;
	}
	bi->bi_magic = BIDR_MAGIC;

	/*
	**	Read the first header to determine the pixel size
	*/

	if (read(bi->bi_fid, (char *)&hdr, sizeof(hdr)) != sizeof(hdr)
			|| strncmp(hdr.h_sfdu, "NJPL1I", 6)) {
		(void)bidr_close((gimage *)bi);
		gipserrno = GIPSBADOPEN;
		return NULL;
	}
	orbit = load_short(hdr.h_orbit);
	if (strncmp(hdr.h_sfdu+8, "0111", 4)) {
		if ((debug & 2) || ORBIT_IS_RIGHT_LOOKING(orbit))
			bi->bi_flag |= BI_OFF4;
		pixsiz = 0.075;
	} else {
		if ((debug & 2) || ORBIT_IS_RIGHT_LOOKING(orbit))
			bi->bi_flag |= BI_OFF1;
		pixsiz = 0.225;
	}

	load_float(olon, hdr.h_olon);
	load_float(olat, hdr.h_olat);
	(void)lseek(bi->bi_fid, (off_t)0, 0);
	bi->bi_xmin = load_long(hdr.h_ref_pixels);
	bi->bi_ymin = load_long(hdr.h_ref_lines);
	bi->bi_xmax = 25000;
	i = 1000000;
	ihput("ynum", (char *)&i);

	/*
	**	Read any BIDRINDX file to save time
	*/

	(void)read_bidr_index(bi, (M_PI*6051)/(pixsiz*180), olat != 0.0);

	/*
	**	Try to read the first data block
	*/

	if (bidr_cell(0, 0, (gimage *)bi) == NULL) {
		(void)bidr_close((gimage *)bi);
		return NULL;
	}

	/*
	**	Construct GIPS "mapping" variable
	*/

	(void)ihput("xnum", (char *)&bi->bi_xmax);
	(void)sprintf(buf, GIPS_MAPPING_P, "sinusoidal",
		16, -(double)(bi->bi_xmin),
		olat ? -(double)(bi->bi_ymin) : (double)(bi->bi_ymin),
		6051/pixsiz, olon, olat, 0.0, 0.0, 0.0, olat ? -1 : 0);
	(void)ihput("mapping", buf);

	/*
	**	Supply header for debug output
	*/

	if (debug & 1) {
		bi->bi_flag |= BI_DEBUG;
		(void)ihget("ynum", (char *)&i, 0);
		(void)fprintf(stderr,
			"file=\"%s\" size=%dx%d pixsiz=%gm\n",
			bi->bi_fname, bi->bi_xmax, i, 1000*pixsiz);
		(void)fprintf(stderr,
			"origin=%d,%d radius=%g euler=%g,%g,0 xyrot=%d\n",
			-bi->bi_xmin, olat ? -bi->bi_ymin : bi->bi_ymin,
			6051/pixsiz, olon, olat, olat ? -1 : 0);
		if (bi->bi_flag & (BI_OFF4 | BI_OFF1))
			(void)fprintf(stderr,
				"orbit=%d, swath offsets corrected\n", orbit);
		(void)fprintf(stderr,
			"\n   Lat      Lon      X     Y      Lseek\n");
	}

	return (gimage *)bi;
}

	/*
	**	Try to read a BIDRINDX file to speed lookups
	*/

static
read_bidr_index(bi, pix, oblique)
	BI		*bi;		/* image block pointer */
	float		pix;		/* pixels per degree */
	int		oblique;	/* =1 if oblique projection */
{
	FILE		*fp;		/* index file pointer */
	char		buf[256];	/* index file name */
	long		lblsize;	/* label length */
	long		nblks;		/* number of BIDR data blocks */
	long		nints;		/* number of 512-byte index blocks */
	off_t		off;		/* offset to BIDR block */
	long		*ibuf;		/* input array */
	long		minxoff;	/* minimum x-address */
	BG		*bg;		/* block pointer */
	register long	*ib;		/* pointer to ibuf */
	register int	n;		/* scratch */
	register char	*c;		/* scratch */

	/*
	**	Try to open any BIDRINDX file
	*/

	if (c = rindex(strcpy(buf, bi->bi_fname), '.'))
		*c = '\0';
	if ((fp = fopen(strcat(buf, ".aux"), "r")) == NULL) {
		(void)strcpy(rindex(buf, '.'), ".AUX");
		if ((fp = fopen(buf, "r")) == NULL)
			return 1;
	}

	/*
	**	Skip over the VICAR label
	*/

	if (fscanf(fp, "LBLSIZE=%d", &lblsize) != 1)
		return fclose(fp);
	(void)fseek(fp, lblsize, 0);

	/*
	**	Read number of data blocks
	*/

	if (fread((char *)&nblks, sizeof(nblks), 1, fp) != 1)
		return fclose(fp);
	nblks = load_long(&nblks);
	minxoff = oblique ? 9999999 : 0;

	/*
	**	Seek to start of needed data
	*/

	nints = (sizeof(long)*nblks+511)/512;
	ibuf = (long *)alloc(512, nints);
	(void)fseek(fp, lblsize+512*(5*nints+1), 0);

	/*
	**	Allocate empty block chain
	*/

	for (n = 0; n < nblks; n++) {
		bg = (BG *)alloc(sizeof(BG), 1);
		if (bi->bi_first == NULL) {
			bi->bi_first = bg;
		} else {
			bg->bg_last = bi->bi_last;
			bi->bi_last->bg_next = bg;
		}
		bi->bi_last = bg;
	}

	/*
	**	Read 5 data types
	*/

	for (n = 0; n < 5; n++) {
		if (fread((char *)(ib = ibuf), 512, (int)nints, fp) != nints)
			break;
		for (off = 0, bg = bi->bi_first; bg; bg = bg->bg_next, ib++)
			switch (n) {
			case 0:
				bg->bg_ynum = load_long(ib);
				if (oblique && bg->bg_last)
					bg->bg_yorg = bg->bg_last->bg_yorg
						+ bg->bg_last->bg_ynum;
				break;
			case 1:
				bg->bg_xnum = load_long(ib);
				off += sizeof(hdr) + bg->bg_xnum * bg->bg_ynum;
				bg->bg_off = off;
				break;
			case 2:
				load_float(bg->bg_lat, ib);
				if (oblique)
					break;
				bg->bg_yorg = 0.5 + pix *
					(bi->bi_first->bg_lat - bg->bg_lat);
				break;
			case 3:
				load_float(bg->bg_lon, ib);
				break;
			case 4:
				bg->bg_xorg = load_long(ib);
				if (bg->bg_xnum + bg->bg_xorg > off)
					off = bg->bg_xnum + bg->bg_xorg;
				if (bg->bg_xorg < minxoff)
					minxoff = bg->bg_xorg;
				break;
			}
	}

	/*
	**	save limits
	*/

	if (ferror(fp)) {
		for (bg = bi->bi_first; bg; bg = bg->bg_next)
			free((char *)bg);
		bi->bi_first = bi->bi_last = NULL;
	} else {
		bi->bi_ptr = bg = bi->bi_first;
		bi->bi_xmin = minxoff;
		bi->bi_xmax = off - bi->bi_xmin;
		if (! oblique)
			bi->bi_ymin = 0.5 + pix * (bg->bg_lat - olat);
		bi->bi_ymax = bi->bi_last->bg_yorg + bi->bi_last->bg_ynum;
		ihput("ynum", (char *)&bi->bi_ymax);
		if (debug & 1)
			(void)fprintf(stderr, "index=\"%s\" nblk=%d xorg=%d\n",
				buf, nblks, bg->bg_xorg - minxoff);
	}
	free((char *)ibuf);
	return fclose(fp);
}

	/*
	**	Close the BIDR
	*/

bidr_close(gim)
	gimage	*gim;				/* input image handle */
{
	BI	*bi;				/* -> image structure */
	BG	*bg;				/* -> block descriptor */

	gipserrno = 0;
	if (gim == NULL)
		return EOF;
	bi = (BI *)gim;
	if (bi->bi_flag & BI_DEBUG) {
		(void)fprintf(stderr, "\n");
		(void)fflush(stderr);
	}
	if (bi->bi_fname)
		free(bi->bi_fname);
	for (bg = bi->bi_first; bg; bg = bg->bg_next)
		free((char *)bg);
	free((char *)bi);
	return close(bi->bi_fid);
}

	/*
	**	Return pointer to pixels (x,y), (x+1,y), ... (x+n-1,y).
	**	If none, return NULL, else _gcount = n.
	*/

char	*
bidr_cell(x, y, gim)
	register int	x;			/* x coordinate */
	register int	y;			/* y coordinate */
	register gimage	*gim;			/* image handle */
{
	register int	size;			/* size of data block */
	register int	p1;			/* offset to start of line */
	register int	p2;			/* offset to end of line */
	register BI	*bi;			/* -> image structure */
	register BG	*bg;			/* -> block descriptor */
	register uchar	*c;			/* -> block pixel */

	gipserrno = 0;
	if (gim == NULL || x < 0 || y < 0)
		return NULL;
	bi = (BI *)gim;
	_gcount = 1024;

	/*
	**	If the pixel is in an image block not yet read,
	**	go read through the BIDR until it is located,
	**	saving a BG structure for each block along the way.
	*/

	while (y > bi->bi_ymax - 1) {
		if (bi->bi_flag & BI_EOF)
			return nulls;
		if (bi->bi_last)
			(void)lseek(bi->bi_fid, bi->bi_last->bg_off, 0);
		if (read(bi->bi_fid, (char *)&hdr, sizeof(hdr)) != sizeof(hdr)
				|| strncmp(hdr.h_sfdu, "NJPL1I", 6)) {
			if ((bi->bi_flag & BI_DEBUG) && hdr.h_sfdu[0] != '^')
				(void)fprintf(stderr,
				"%s: bad SFDU label: '%-12.12s' at offset %d\n",
					bi->bi_fname, hdr.h_sfdu,
					bi->bi_last ? bi->bi_last->bg_off : 0);
			if (errno)
				gipserrno = GIPSBADREAD;
			bi->bi_flag |= BI_EOF;
			return nulls;
		}
		bg = bi->bi_ptr = (BG *)alloc(sizeof(BG), 1);
		bg->bg_xorg = load_long(hdr.h_ref_pixels);
		bg->bg_yorg = load_long(hdr.h_ref_lines);
		bg->bg_xnum = load_short(hdr.h_pixels);
		bg->bg_ynum = load_short(hdr.h_lines);
		load_float(bg->bg_lat, hdr.h_rlat);
		load_float(bg->bg_lon, hdr.h_rlon);

		if (bi->bi_first == NULL) {
			bi->bi_first = bg;
			bi->bi_xmin = bg->bg_xorg;
			bi->bi_ymin = bg->bg_yorg;
		} else {
			bg->bg_last = bi->bi_last;
			bi->bi_last->bg_next = bg;
			bg->bg_off = bg->bg_last->bg_off;
		}
		bi->bi_last = bg;
		bg->bg_off += sizeof(hdr) + bg->bg_xnum * bg->bg_ynum;
		bg->bg_yorg = abs(bi->bi_ymin - bg->bg_yorg);
		bi->bi_ymax = bg->bg_yorg + bg->bg_ynum;
	}

	/*
	**	Advance up/down BG chain until correct block is found
	*/

	bg = bi->bi_ptr;

	while (y < bg->bg_yorg)
		if (bg->bg_last)
			bi->bi_ptr = bg = bg->bg_last;
		else
			return nulls;

	while (y > bg->bg_yorg + bg->bg_ynum - 1)
		bi->bi_ptr = bg = bg->bg_next;

	/*
	**	If the image block isn't in core, go read it
	*/

	if (bg != bi->bi_save) {
		size = bg->bg_xnum * bg->bg_ynum;
		if (size > bi->bi_size) {
			if (bi->bi_buf)
				free(bi->bi_buf);
			bi->bi_buf = alloc(size, 1);
			bi->bi_size = size;
		}
		(void)lseek(bi->bi_fid, (off_t)(bg->bg_off - size), 0);
		if (read(bi->bi_fid, bi->bi_buf, size) != size)
			return nulls;
		bi->bi_save = bg;
		if (bi->bi_flag & BI_DEBUG)
			(void)fprintf(stderr, "\r%8.3f %8.3f %6d %5d %9d",
				bg->bg_lat, bg->bg_lon, bg->bg_xorg,
				bg->bg_yorg, bg->bg_off);
		bi->bi_p1 = load_short(bi->bi_buf);
		bi->bi_p2 = load_short(bi->bi_buf+2);
	}

	if ((y -= bg->bg_yorg) < 0 || y >= bg->bg_ynum)
		return nulls;

	/*
	**	Point to the desired raster
	*/

	c = (uchar *)(bi->bi_buf + y * bg->bg_xnum);
	size = (bi->bi_flag & BI_OFF4) ? 0 : 4;

	/*
	**	Determine the "good" swath pointers: p1,p2
	*/

	if (y) {
		p1 = load_short(c) + size;
		p2 = load_short(c+2) + size;
	} else if (bg->bg_ynum != 1) {
		p1 = load_short(c+bg->bg_xnum) + size;
		p2 = load_short(c+bg->bg_xnum+2) + size;
	} else {
		p1 = 4;
		p2 = bg->bg_xnum;
	}

	if (bi->bi_flag & BI_OFF1)
		--p2;

	/*
	**	In debug mode, outline the "good" swath
	*/

	if (bi->bi_flag & BI_DEBUG) {
		if (p1 > 4 && p1 <= bg->bg_xnum)
			c[p1-1] = 255;
		if (p2 > 4 && p2 < bg->bg_xnum)
			c[p2] = 255;
		p1 = 4;
		p2 = bg->bg_xnum;
	}

	x += bi->bi_xmin - bg->bg_xorg + 4;

	if (x >= p2 || p1 >= p2)
		return nulls;

	if (x < p1) {
		_gcount = p1 - x;
		if (_gcount > 1024)
			_gcount = 1024;
		return nulls;
	}

	/*
	**	Locate the specific pixel
	*/

	_gcount = p2 - x;
	return (char *)(c + x);
}
	/*
	**	Interpolate cell value
	*/
char *
bidr_interp(vx, vy, gim)
	float		vx, vy;
	gimage		*gim;
{
	register BI	*bi;
	register	x = vx, y = vy, i;
	register char	*c;
	static char	ch;
	int		n[4];
	float		vxvy;

	if (! (bi = (BI *)gim))
		return NULL;

	n[0] = *(c = bidr_cell(x,y,bi)) & 255;
	n[1] = ((_gcount > 1) ? *++c : *bidr_cell(x+1,y,bi)) & 255;
	n[2] = *(c = bidr_cell(x,y+1,bi)) & 255;
	n[3] = ((_gcount > 1) ? *++c : *bidr_cell(x+1,y+1,bi)) & 255;
	vx -= x;
	vy -= y;
	vxvy = vx*vy;
	for (i = 0; i < 4 && ! n[i]; i++)
		;
	if (i == 4)
		ch = 0;
	else
		ch = (1-vx-vy+vxvy)	* (n[0] ? n[0] : n[i])
			+ (vx-vxvy)	* (n[1] ? n[1] : n[i])
			+ (vy-vxvy)	* (n[2] ? n[2] : n[i])
			+ vxvy		* (n[3] ? n[3] : n[i]);
	return &ch;
}
